# -*- coding: utf-8 -*-

import os
import torch as th
import leibniz as lbnz

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.animation import FuncAnimation, writers

from leibniz.diffeq import odeint as odeint


def binary(tensor):
    return th.where(tensor > lbnz.zero, lbnz.one, lbnz.zero)


def draw_3D_plot(x, y, fld, fig_name='test.png'):
    x = x.cpu().numpy()
    y = y.cpu().numpy()
    fld = fld.cpu().numpy()

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    c = ax.plot_surface(x, y, fld, rstride=1, cstride=1,
                        cmap='binary', edgecolor='none', vmin=-2, vmax=2)
    ax.set_xlim(-16, 16)
    ax.set_ylim(-16, 16)
    ax.set_zlim(-2, 2)

    fig.colorbar(c)
    plt.savefig(os.path.join('results', fig_name))
    plt.close()


def draw_animation(sequence, x, y, gif_name='test.gif', frames=51, mode='3d'):
    gif_name = os.path.join('results', gif_name)
    x = x.cpu().numpy()
    y = y.cpu().numpy()
    sequence = [item.cpu().numpy() for item in sequence]

    fig = plt.figure()

    def update(n):
        print('drawing %2dth image in animation' % n)
        plt.clf()
        ax = fig.gca(projection='3d')
        c = ax.plot_surface(x, y, sequence[n], rstride=1, cstride=1,
                            cmap='binary', edgecolor='none', vmin=-2, vmax=2)
        ax.set_xlim(-16, 16)
        ax.set_ylim(-16, 16)
        ax.set_zlim(-2, 2)
        fig.colorbar(c)

    ani = FuncAnimation(fig, update, frames=frames, interval=1000.0 / 4, save_count=100, repeat_delay=1000, repeat=True)

    # save it into gif (need imagemagick)
    ani.save(gif_name[:-4] + mode + '.gif', writer='imagemagick')
    plt.close()


if __name__ == '__main__':
    from leibniz.core3d.gridsys.regular3 import RegularGrid

    lbnz.bind(RegularGrid(
        basis='x,y,z',
        W=51, L=51, H=15,
        east=16.0, west=-16.0,
        north=16.0, south=-16.0,
        upper=6.0, lower=1.0
    ))
    lbnz.use('x,y,z')

    dt = 1 / 120
    t_iter = 1000

    init = th.exp(-(((lbnz.x - 1) * 2) ** 2 + ((lbnz.y - 1) * 2) ** 2))  # Gaussian

    r = th.sqrt(lbnz.x ** 2 + lbnz.y ** 2)
    theta = th.atan2(lbnz.y, lbnz.x)
    wind = - r * th.sin(theta), r * th.cos(theta), lbnz.zero

    def derivitive(t, clouds):
        return - lbnz.upwind(wind, clouds)

    pred = odeint(derivitive, init, th.arange(0, 8, 1 / 200), method='rk4')
    sequence = [pred[i].squeeze()[..., lbnz.H // 2] for i in range(1600) if i % 20 == 0]
    draw_animation(sequence, lbnz.x.squeeze()[..., 0], lbnz.y.squeeze()[..., 0], 'simple_rotation.gif',
                   mode='3d', frames=len(sequence))



# -*- coding: utf-8 -*-


_name_ = 'x,y'
_params_ = ('x', 'y')


def transform(x, y, **kwargs):
    return x, y


def dtransform(x, y, dx, dy, **kwargs):
    return dx, dy


def inverse(x, y, **kwargs):
    return x, y


def dinverse(x, y, dx, dy, **kwargs):
    return dx, dy




# -*- coding: utf-8 -*-

import torch as th
import torch.nn as nn


class Basic(th.nn.Module):
    extension = 1
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv):
        super(Basic, self).__init__()

        self.step = step
        self.relu = relu
        self.conv1 = conv(dim, dim, kernel_size=3, stride=1, padding=1, groups=1, bias=False, dilation=1)
        self.conv2 = conv(dim, dim, kernel_size=3, stride=1, padding=1, groups=1, bias=False, dilation=1)

        nn.init.normal_(self.conv1.weight, 0.0, 0.04)
        nn.init.normal_(self.conv2.weight, 0.0, 0.04)

    def forward(self, x):

        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = x + y

        return y


class Bottleneck(th.nn.Module):
    extension = 1
    least_required_dim = 4

    def __init__(self, dim, step, relu, conv):
        super(Bottleneck, self).__init__()

        self.step = step
        self.relu = relu
        self.conv1 = conv(dim, dim // 4, kernel_size=1, bias=False)
        self.conv2 = conv(dim // 4, dim // 4, kernel_size=3, bias=False, padding=1)
        self.conv3 = conv(dim // 4, dim, kernel_size=1, bias=False)

        nn.init.normal_(self.conv1.weight, 0.0, 0.04)
        nn.init.normal_(self.conv2.weight, 0.0, 0.04)
        nn.init.normal_(self.conv3.weight, 0.0, 0.04)

    def forward(self, x):

        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.relu(y)
        y = self.conv3(y)
        y = x + y

        return y



from torch import nn as nn


def conv7x7(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """7x7 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=7, stride=stride,
                     padding=3, groups=groups, bias=False, dilation=dilation)


def conv5x5(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """5x5 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=5, stride=stride,
                     padding=2, groups=groups, bias=False, dilation=dilation)


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)



# -*- coding: utf-8 -*-

import numpy as np
import torch as th
import logging

from typing import Tuple
from torch import Tensor
from torch.nn.functional import conv2d, conv3d

default_device = 0
used_modules = []
used_environ = {}

logger = logging.getLogger('leibniz')

lanczosk2d = None
lanczosk3d = None


def get_device():
    return default_device


def set_device(ix):
    global default_device, lanczosk2d, lanczosk3d

    if default_device != -1 and th.cuda.is_available():
        if lanczosk2d is not None:
            lanczosk2d = lanczosk2d.cuda(device=ix)

        if lanczosk3d is not None:
            lanczosk3d = lanczosk3d.cuda(device=ix)

    if default_device == ix:
        return

    logger.info('reset device from %d to %d', default_device, ix)

    if _grid_ is not None:
        _grid_.set_device(ix)
    if _frame_ is not None:
        _frame_.set_device(ix)
    if _element_ is not None:
        _element_.set_device(ix)

    if default_device != -1 and ix == -1:
        if lanczosk2d is not None:
            lanczosk2d = lanczosk2d.cpu()

        if lanczosk3d is not None:
            lanczosk3d = lanczosk3d.cpu()

    default_device = ix

    for mname in used_modules:
        use(mname, **used_environ[mname])

    globals().update({
        'L': _grid_.L,
        'W': _grid_.W,
        'H': _grid_.H,
        'boundary': lambda: _grid_.boundary(),
        'mk_zero': _grid_.mk_zero,
        'mk_one': _grid_.mk_one,
        'zero': _grid_.zero,
        'one': _grid_.one,
        'random': lambda: _grid_.random(),
        'dL': _element_.dL,
        'dS': _element_.dS,
        'dV': _element_.dV,
    })


def cast(element, device=-1) -> Tensor:
    element = np.array(element, dtype=np.float64)
    tensor = th.DoubleTensor(element)
    if device != -1 and th.cuda.is_available():
        return tensor.cuda(device=device)
    else:
        return tensor


diff = None
upwindp = None
upwindm = None

_grid_ = None
_frame_ = None
_element_ = None
_ops_ = None
_v3_ = None


def clear():
    global _grid_, _frame_, _element_, _ops_, _v3_
    _grid_ = None
    _frame_ = None
    _element_ = None
    _ops_ = None
    _v3_ = None

    used_modules.clear()
    used_environ.clear()


def load(mname):
    import importlib

    if ',' in mname:
        return importlib.import_module('leibniz.core3d.basis.%s' % mname.replace(',', '_'))
    else:
        return importlib.import_module('leibniz.core3d.frame.%s' % mname)


def use(mname, **kwargs):
    if mname not in used_modules:
        used_modules.append(mname)
        used_environ[mname] = kwargs

    if ',' not in mname:
        target = load(mname)
        pname = target._name_
        glbs = globals()
        pval = target._clazz_(_grid_)
        glbs.update({
            pname: pval,
        })
    else:
        data = _grid_.data()
        delta = _grid_.delta()

        if mname == _grid_.basis:
            target = load(mname)
            pvalues = data[:, 0:1], data[:, 1:2], data[:, 2:3]
            dvalues = delta[:, 0:1], delta[:, 1:2], delta[:, 2:3]
        else:
            origin = load(_grid_.basis)
            target = load(mname)

            tval1, tval2, tval3 = origin.transform(data[:, 0:1], data[:, 1:2], data[:, 2:3], **kwargs)
            pval1, pval2, pval3 = target.inverse(tval1, tval2, tval3, **kwargs)
            mval1, mval2, mval3 = origin.dtransform(
                data[:, 0:1], data[:, 1:2], data[:, 2:3],
                delta[:, 0:1], delta[:, 1:2], delta[:, 2:3], **kwargs)
            dval1, dval2, dval3 = target.dinverse(tval1, tval2, tval3, mval1, mval2, mval3, **kwargs)

            pvalues = pval1, pval2, pval3
            dvalues = dval1, dval2, dval3

        for ix, pname in enumerate(target._params_):
            dname = 'd%s' % pname
            pval = pvalues[ix]
            dval = dvalues[ix]

            globals().update({
                pname: pval,
                dname: dval
            })


def bind(grid, **kwargs):
    if 'device' in kwargs.keys():
        device_to = kwargs['device']
    else:
        device_to = default_device

    logger.info('device_to: %d', device_to)

    import leibniz.core3d.vec3 as v3
    import leibniz.core3d.diffr.regular3 as d

    from leibniz.core3d.element import Elements
    from leibniz.core3d.ops import LocalOps

    global _grid_, _frame_, _element_, _ops_, _v3_, diff, upwindm, upwindp

    _grid_ = grid
    diff = d.central5
    upwindp = d.upwind_p4
    upwindm = d.upwind_m4

    globals().update({
        'L': _grid_.L,
        'W': _grid_.W,
        'H': _grid_.H,
        'boundary': _grid_.boundary(),
        'zero': _grid_.zero,
        'one': _grid_.one,
        'random': _grid_.random(),
    })

    use(_grid_.basis, **kwargs)

    _element_ = Elements(**kwargs)
    _ops_ = LocalOps(_grid_, _element_, diff)
    _v3_ = v3

    globals().update({
        'dL': _element_.dL,
        'dS': _element_.dS,
        'dV': _element_.dV,
    })

    set_device(device_to)
    d.set_device(device_to)


def dot(x: Tuple[Tensor, Tensor, Tensor], y: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.dot(x, y)


def cross(x: Tuple[Tensor, Tensor, Tensor], y: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.cross(x, y)


def norm(v: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.norm(v)


def normsq(v: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.normsq(v)


def normalize(v: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
    return _v3_.normalize(v)


def box(a: Tuple[Tensor, Tensor, Tensor], b: Tuple[Tensor, Tensor, Tensor], c: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.box(a, b, c)


def det(transform: Tensor) -> Tensor:
    return _v3_.det(transform)


def grad(scalar: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.grad(scalar)


def div(vector: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _ops_.div(vector)


def curl(vector: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.curl(vector)


def laplacian(scalar: Tensor) -> Tensor:
    return _ops_.laplacian(scalar)


def vlaplacian(vector: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.vlaplacian(vector)


def adv(wind: Tuple[Tensor, Tensor, Tensor], scalar: Tensor, filter=None) -> Tensor:
    return _ops_.adv(wind, scalar, filter)


def vadv(wind: Tuple[Tensor, Tensor, Tensor], vector: Tuple[Tensor, Tensor, Tensor], filter=None) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.vadv(wind, vector, filter)


def upwind(wind: Tuple[Tensor, Tensor, Tensor], scalar: Tensor, filter=None) -> Tensor:
    return _ops_.upwind(wind, scalar, filter, upwindp=upwindp, upwindm=upwindm)


avgk = cast([[[
    [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
    [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
    [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
]]]) / 27.0


def avgfilter(f):
    return conv3d(f, avgk, padding=1)


lanczosk2d = cast(np.sinc([[[
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [-3, -2, -1, 0, +1, +2, +3],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
]]]) * np.sinc([[[
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [-3 / 3, -2 / 3, -1 / 3, 0, +1 / 3, +2 / 3, +3 / 3],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
]]]))


lanczosk2d = lanczosk2d * lanczosk2d.transpose(-2, -1)


def lanczosfilter(f):
    return conv2d(f, lanczosk2d, padding=3)


lanczosk3d = cast(np.sinc([[[[
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [-3, -2, -1, 0, +1, +2, +3],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
]]]]) * np.sinc([[[[
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [-3 / 3, -2 / 3, -1 / 3, 0, +1 / 3, +2 / 3, +3 / 3],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0],
]]]]))


lanczosk3d = lanczosk3d * lanczosk3d.transpose(-2, -1) * lanczosk3d.transpose(-3, -1)


def lanczos3dfilter(f):
    return conv3d(f, lanczosk3d, padding=3)





import torch
from .misc import _convert_to_tensor, _dot_product


def _interp_fit(y0, y1, y_mid, f0, f1, dt):
    """Fit coefficients for 4th order polynomial interpolation.

    Args:
        y0: function value at the start of the interval.
        y1: function value at the end of the interval.
        y_mid: function value at the mid-point of the interval.
        f0: derivative value at the start of the interval.
        f1: derivative value at the end of the interval.
        dt: width of the interval.

    Returns:
        List of coefficients `[a, b, c, d, e]` for interpolating with the polynomial
        `p = a * x ** 4 + b * x ** 3 + c * x ** 2 + d * x + e` for values of `x`
        between 0 (start of interval) and 1 (end of interval).
    """
    a = tuple(
        _dot_product([-2 * dt, 2 * dt, -8, -8, 16], [f0_, f1_, y0_, y1_, y_mid_])
        for f0_, f1_, y0_, y1_, y_mid_ in zip(f0, f1, y0, y1, y_mid)
    )
    b = tuple(
        _dot_product([5 * dt, -3 * dt, 18, 14, -32], [f0_, f1_, y0_, y1_, y_mid_])
        for f0_, f1_, y0_, y1_, y_mid_ in zip(f0, f1, y0, y1, y_mid)
    )
    c = tuple(
        _dot_product([-4 * dt, dt, -11, -5, 16], [f0_, f1_, y0_, y1_, y_mid_])
        for f0_, f1_, y0_, y1_, y_mid_ in zip(f0, f1, y0, y1, y_mid)
    )
    d = tuple(dt * f0_ for f0_ in f0)
    e = y0
    return [a, b, c, d, e]


def _interp_evaluate(coefficients, t0, t1, t):
    """Evaluate polynomial interpolation at the given time point.

    Args:
        coefficients: list of Tensor coefficients as created by `interp_fit`.
        t0: scalar float64 Tensor giving the start of the interval.
        t1: scalar float64 Tensor giving the end of the interval.
        t: scalar float64 Tensor giving the desired interpolation point.

    Returns:
        Polynomial interpolation of the coefficients at time `t`.
    """

    dtype = coefficients[0][0].dtype
    device = coefficients[0][0].device

    t0 = _convert_to_tensor(t0, dtype=dtype, device=device)
    t1 = _convert_to_tensor(t1, dtype=dtype, device=device)
    t = _convert_to_tensor(t, dtype=dtype, device=device)

    assert (t0 <= t) & (t <= t1), 'invalid interpolation, fails `t0 <= t <= t1`: {}, {}, {}'.format(t0, t, t1)
    x = ((t - t0) / (t1 - t0)).type(dtype).to(device)

    xs = [torch.tensor(1).type(dtype).to(device), x]
    for _ in range(2, len(coefficients)):
        xs.append(xs[-1] * x)

    return tuple(_dot_product(coefficients_, reversed(xs)) for coefficients_ in zip(*coefficients))



# -*- coding: utf-8 -*-

import numpy as np
import torch as th
import torch.nn as nn

from leibniz.unet.senet import SELayer


class HyperBasic(nn.Module):
    extension = 2
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv, reduction=16):
        super(HyperBasic, self).__init__()
        self.dim = dim
        self.step = step

        if relu is None:
            self.relu = nn.ReLU(inplace=True)
        else:
            self.relu = relu

        if conv is None:
            self.conv = nn.Conv2d
        else:
            self.conv = conv

        self.conv0 = self.conv(dim, dim, kernel_size=3, padding=1)
        self.conv1 = self.conv(dim, dim // 2, kernel_size=3, padding=1)
        self.conv2 = self.conv(3 * dim, dim, kernel_size=3, padding=1)
        self.conv3 = self.conv(dim, dim, kernel_size=3, padding=1)
        self.se = SELayer(dim, reduction)

    def forward(self, y):
        x, theta = y[:, 0:self.dim // 2], y[:, self.dim // 2:self.dim]
        u = self.conv0(y)
        u = self.relu(u)
        u = self.conv1(u)

        cs = self.step * u * th.cos(theta * np.pi * 6)
        ss = self.step * u * th.sin(theta * np.pi * 6)

        y1 = (1 + ss) * x + cs
        y2 = (1 + cs) * x - ss
        y3 = (1 - cs) * x + ss
        y4 = (1 - ss) * x - cs
        ys = th.cat((y1, y2, y3, y4, cs, ss), dim=1)

        dy = self.conv2(ys)
        dy = self.relu(dy)
        dy = self.conv3(dy)

        y = y + self.se(dy)

        return y


class HyperBottleneck(nn.Module):
    extension = 4
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv, reduction=16):
        super(HyperBottleneck, self).__init__()
        self.dim = dim
        self.step = step

        if relu is None:
            self.relu = nn.ReLU(inplace=True)
        else:
            self.relu = relu

        if conv is None:
            self.conv = nn.Conv2d
        else:
            self.conv = conv

        self.conv0 = self.conv(dim, dim, kernel_size=3, padding=1)
        self.conv1 = self.conv(dim, dim // 2, kernel_size=3, padding=1)
        self.conv2 = self.conv(3 * dim, 3 * dim // 4, kernel_size=1, bias=False)
        self.conv3 = self.conv(3 * dim // 4, dim // 2, kernel_size=3, bias=False, padding=1)
        self.conv4 = self.conv(dim // 2, dim, kernel_size=1, bias=False)
        self.se = SELayer(dim, reduction)

    def forward(self, y):

        x, theta = y[:, 0:self.dim // 2], y[:, self.dim // 2:self.dim]
        u = self.conv0(y)
        u = self.relu(u)
        u = self.conv1(u)

        cs = self.step * u * th.cos(theta * np.pi * 6)
        ss = self.step * u * th.sin(theta * np.pi * 6)

        y1 = (1 + ss) * x + cs
        y2 = (1 + cs) * x - ss
        y3 = (1 - cs) * x + ss
        y4 = (1 - ss) * x - cs
        ys = th.cat((y1, y2, y3, y4, cs, ss), dim=1)

        dy = self.conv2(ys)
        dy = self.relu(dy)
        dy = self.conv3(dy)
        dy = self.relu(dy)
        dy = self.conv4(dy)

        y = y + self.se(dy)

        return y



# -*- coding: utf-8 -*-

import numpy as np
import torch as th
import torch.nn as nn

from leibniz.unet.senet import SELayer


class CmplxHyperBasic(nn.Module):
    extension = 2
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv, reduction=16):
        super(CmplxHyperBasic, self).__init__()
        self.dim = dim
        self.step = step

        if relu is None:
            self.relu = nn.ReLU(inplace=True)
        else:
            self.relu = relu

        if conv is None:
            self.conv = nn.Conv2d
        else:
            self.conv = conv

        self.conv0 = self.conv(dim, dim, kernel_size=3, padding=1)
        self.conv1 = self.conv(dim, dim // 2, kernel_size=3, padding=1)
        self.conv2 = self.conv(dim, dim, kernel_size=3, padding=1)
        self.conv3 = self.conv(dim, dim // 2, kernel_size=3, padding=1)
        self.se = SELayer(dim, reduction)

    def forward(self, z):
        x, y = z[:, 0:self.dim // 2], z[:, self.dim // 2:self.dim]

        u = self.conv0(z)
        u = self.relu(u)
        u = self.conv1(u)

        theta = self.conv2(z)
        theta = self.relu(theta)
        theta = self.conv3(theta)

        cs = self.step * u * th.cos(theta * np.pi * 6)
        ss = self.step * u * th.sin(theta * np.pi * 6)

        dx = ss * x - cs * y + cs
        dy = ss * y + cs * x + ss
        dz = th.cat([dx, dy], dim=1)

        z = z + self.se(dz)

        return z


class CmplxHyperBottleneck(nn.Module):
    extension = 6
    least_required_dim = 1

    def __init__(self, dim, step, relu, conv, reduction=16):
        super(CmplxHyperBottleneck, self).__init__()
        self.dim = dim
        self.step = step

        if relu is None:
            self.relu = nn.ReLU(inplace=True)
        else:
            self.relu = relu

        if conv is None:
            self.conv = nn.Conv2d
        else:
            self.conv = conv

        self.conv0 = self.conv(dim, dim, kernel_size=3, padding=1)
        self.conv1 = self.conv(dim, dim // 3, kernel_size=3, padding=1)
        self.conv2 = self.conv(10 * dim // 3, 10 * dim // 12, kernel_size=1, bias=False)
        self.conv3 = self.conv(10 * dim // 12, dim // 3, kernel_size=3, bias=False, padding=1)
        self.conv4 = self.conv(dim // 3, dim, kernel_size=1, bias=False)
        self.se = SELayer(dim, reduction)

    def forward(self, z):
        x, y, theta = z[:, 0:self.dim // 3], z[:, self.dim // 3:self.dim * 2 // 3], z[:, self.dim * 2 // 3:self.dim]
        u = self.conv0(z)
        u = self.relu(u)
        u = self.conv1(u)

        cs = self.step * u * th.cos(theta * np.pi * 6)
        ss = self.step * u * th.sin(theta * np.pi * 6)

        x1 = (1 + ss) * x - cs * y + cs
        y1 = (1 + ss) * y + cs * x + ss
        x2 = (1 + cs) * x + ss * y - ss
        y2 = (1 + cs) * y - ss * x + cs
        x3 = (1 - ss) * x + cs * y - cs
        y3 = (1 - ss) * y - cs * x - ss
        x4 = (1 - cs) * x - ss * y + ss
        y4 = (1 - cs) * y + ss * x - cs
        zs = th.cat((x1, y1, x2, y2, x3, y3, x4, y4, cs, ss), dim=1)

        dz = self.conv2(zs)
        dz = self.relu(dz)
        dz = self.conv3(dz)
        dz = self.relu(dz)
        dz = self.conv4(dz)

        z = z + self.se(dz)

        return z



# -*- coding: utf-8 -*-

import torch as th
import torch.nn as nn


class Swish(th.nn.Module):
    def __init__(self):
        super(Swish, self).__init__()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        return x * self.sigmoid(x)


class Mish(th.nn.Module):
    def __init__(self):
        super(Mish, self).__init__()
        self.tanh = nn.Tanh()
        self.softplus = nn.Softplus()

    def forward(self, x):
        return x * self.tanh(self.softplus(x))


class CappingRelu(th.nn.Module):
    def __init__(self):
        super(CappingRelu, self).__init__()
        self.leaky = nn.LeakyReLU(inplace=True)

    def forward(self, x):
        return th.clamp(self.leaky(x), max=10)


class Atanh(nn.Module):
    def __init__(self):
        super(Atanh, self).__init__()

    def forward(self, x):
        return th.log1p(2 * x / (1 - x)) / 2



# -*- coding: utf-8 -*-

from leibniz.unet.base import UNet
from leibniz.unet.residual import Basic
from leibniz.nn.activation import Swish


def unet4(in_channels, out_channels, spatial=(256, 256)):
    return UNet(in_channels, out_channels, block=None, relu=Swish(), layers=4, ratio=0,
                 vblks=[0, 0, 0, 0], hblks=[0, 0, 0, 0],
                 scales=[-1, -1, -1, -1], factors=[1, 1, 1, 1], spatial=spatial)


def unet8(in_channels, out_channels, spatial=(256, 256)):
    return UNet(in_channels, out_channels, block=None, relu=Swish(), layers=8, ratio=0,
                 vblks=[0, 0, 0, 0, 0, 0, 0, 0], hblks=[0, 0, 0, 0, 0, 0, 0, 0],
                 scales=[-1, -1, -1, -1, -1, -1, -1, -1], factors=[1, 1, 1, 1, 1, 1, 1, 1], spatial=spatial)


def resunet(in_channels, out_channels, block=Basic, relu=Swish(), layers=4, ratio=0,
                 vblks=[1, 1, 1, 1], hblks=[1, 1, 1, 1],
                 scales=[-1, -1, -1, -1], factors=[1, 1, 1, 1], spatial=(256, 256), normalizor='batch'):
    return UNet(in_channels, out_channels, block=block, relu=relu, layers=layers, ratio=ratio,
                 vblks=vblks, hblks=hblks, scales=scales, factors=factors, spatial=spatial, normalizor=normalizor)
